from iterwrapper import IterWrapper
from pydantic.main import ModelMetaclass

from .dispatcher import BaseDispatcher
import copy

class EventMeta(ModelMetaclass):
    class Config:
        arbitrary_types_allowed = True

    def __new__(mcls, name, bases, mapping, **kwargs):
        if any(IterWrapper(bases).filter(lambda x: getattr(x, "__base_event__", False)).collect(list)):
            if not mapping.__contains__("__base_event__"):
                mapping["__base_event__"] = False
        if not mapping.get("Dispatcher") and name != "BaseEvent":
            raise AttributeError("a event class must have a dispatcher called 'Dispatcher'")
        if not mapping.get("Config") or not getattr(mapping.get("Config"), "arbitrary_types_allowed", False):
            if not mapping.get("Config"):
                mapping['Config'] = copy.copy(mcls.Config)
            else:
                mapping['Config'].arbitrary_types_allowed = True
        r = super().__new__(mcls, name, bases, mapping, **kwargs)
        if mapping.get("type"):
            r.type = mapping.get("type")
        return r

class BaseEvent(metaclass=EventMeta):
    Dispatcher: BaseDispatcher

    def __init__(self, **kwargs):
        super().__init__(**kwargs)


from pydantic import BaseModel  # pylint: disable=no-name-in-module

from ..entities.dispatcher import BaseDispatcher
from ..entities.event import BaseEvent
from ..interfaces.dispatcher import DispatcherInterface


class ExceptionThrowed(BaseEvent, BaseModel):
    exception: Exception
    event: BaseEvent

    class Dispatcher(BaseDispatcher):
        @staticmethod
        def catch(interface: DispatcherInterface):
            if interface.annotation == interface.event.exception.__class__:
                return interface.event.exception
            if interface.name == "event":
                return interface.event.event



from abc import ABCMeta, abstractstaticmethod
from typing import List

class BaseDispatcher(metaclass=ABCMeta):
    always: bool
    mixin: List["BaseDispatcher"]

    @abstractstaticmethod
    def catch(interface: "DispatcherInterface"):
        pass

    @staticmethod
    def check(interface: "DispatcherInterface"):
        return True



from typing import Any

from pydantic import BaseModel  # pylint: disable=no-name-in-module


class ObjectContainer(BaseModel):
    target: Any

    def __init__(self, content: Any = None):
        if content.__class__ is self.__class__:
            content = content.target
        super().__init__(target=content)

class Force(ObjectContainer):
    """用于转义在本框架中特殊部分的特殊值

    例如：Dispatcher 返回时 None 表示本级 dispatcher 无法满足需求,
    DispatcherInterface 会继续向下查询,
    而某些时候我们确实是需要传递 None 的，
    这时候可以用本标识来保证 None 被顺利作为一个参数传入。
    """

class RemoveMe(ObjectContainer):
    """当本标识的实例为一受 Executor 影响的 Listener 返回值时,
    Executor 会尝试在当前 Broadcast 实例中找出并删除本 Listener 实例.
    """
    


from typing import Any, Callable, List

from pydantic import BaseModel  # pylint: disable=no-name-in-module

from ..entities.dispatcher import BaseDispatcher
from ..entities.signatures import Force
from ..interfaces.dispatcher import DispatcherInterface


class MappingRule(BaseModel):
    mode: Callable[[DispatcherInterface], bool]
    value: Any

    @classmethod
    def nameEquals(cls, name, value):
        return cls(mode=lambda x: x.name == name, value=value)

    @classmethod
    def annotationEquals(cls, annotation, value):
        return cls(mode=lambda x: x.annotation == annotation, value=value)

    @classmethod
    def defaultEquals(cls, default, value):
        return cls(mode=lambda x: x.default == default, value=value)

def SimpleMapping(rules: List[MappingRule]):
    class mapping_dispatcher(BaseDispatcher):
        @staticmethod
        def catch(interface: DispatcherInterface):
            for rule in rules:
                if rule.mode(interface):
                    return Force(rule.value)
    return mapping_dispatcher

def Hook(condition, fixer):
    class hook_dispatcher(BaseDispatcher):
        @staticmethod
        def catch(interface: DispatcherInterface):
            if condition(interface):
                return fixer(interface.execute_with(
                    interface.name,
                    interface.annotation,
                    interface.default
                ))
    return hook_dispatcher


import inspect
from typing import Any, Dict

from ..entities.decorater import Decorater
from ..entities.signatures import Force
from ..utilles import is_asyncgener, iscoroutinefunction, isgeneratorfunction, run_always_await
from .dispatcher import DispatcherInterface

class DecoraterInterface:
    """Graia Broadcast Control 内部机制 Decorate 的具体管理实现
    """
    dispatcher_interface: DispatcherInterface
    local_storage: Dict[Any, Any] = {}
    return_value: Any = None
    default = None

    def __init__(self, dispatcher_interface: DispatcherInterface):
        self.dispatcher_interface = dispatcher_interface
        self.dispatcher_interface.dispatchers.insert(0, self.decorater_dispatcher)

    @property
    def name(self):
        return self.dispatcher_interface.name

    @property
    def annotation(self):
        return self.dispatcher_interface.annotation

    @property
    def event(self):
        return self.dispatcher_interface.event

    async def decorater_dispatcher(self, interface: DispatcherInterface):
        if isinstance(interface.default, Decorater):
            decorater: Decorater = interface.default
            if not decorater.pre:
                # 作为 装饰
                self.return_value = await interface.execute_with(interface.name, interface.annotation, None)
            try:
                # 这里隐式的复用了 dispatcher interface 的生成器终结者机制
                if is_asyncgener(decorater.target):
                    # 如果是异步生成器
                    async for i in decorater.target(self):
                        yield i
                elif (isgeneratorfunction(decorater.target) and \
                    not iscoroutinefunction(decorater.target)):
                    # 同步生成器
                    for i in decorater.target(self):
                        yield i
                else:
                    yield Force(await run_always_await(decorater.target(self)))
            finally:
                if not decorater.pre:
                    self.return_value = None


from typing import Any, Callable, List, Type, Union

from pydantic.dataclasses import dataclass

from ..entities.dispatcher import BaseDispatcher
from ..entities.event import BaseEvent
from ..entities.listener import Listener


class ExecutorProtocol:
    def __init__(self,
        target: Union[Callable, Listener],
        event: BaseEvent,
        dispatchers: List[Union[
            Type[BaseDispatcher],
            Callable,
            BaseDispatcher,
        ]] = None,
        hasReferrer: bool = False,
        enableInternalAccess: bool = False,
    ) -> None:
        self.target = target
        self.dispatchers = dispatchers or []
        self.event = event
        self.hasReferrer = hasReferrer
        self.enableInternalAccess = enableInternalAccess

    target: Union[Callable, Listener]
    event: BaseEvent
    dispatchers: List[Union[
        Type[BaseDispatcher],
        Callable,
        BaseDispatcher
    ]]
    hasReferrer: bool = False
    enableInternalAccess: bool = False



from graia.broadcast.utilles import isasyncgen, iscoroutinefunction, isgenerator
from ..entities.decorater import Decorater
from typing import Callable, ContextManager, Any, Hashable
from ..entities.signatures import Force
from ..interfaces.decorater import DecoraterInterface
import inspect
from ..exceptions import InvaildContextTarget

class Depend(Decorater):
    pre = True
    depend_callable: Callable
    cache: bool = False

    def __init__(self, callable, *, cache=False):
        self.cache = cache
        self.depend_callable = callable

    def __repr__(self) -> str:
        return "<Depend target={0}>".format(self.depend_callable)

    async def target(self, interface: DecoraterInterface):
        if self.cache:
            attempt = interface.local_storage.get(self.depend_callable)
            if attempt:
                yield Force(attempt)
                return
        result = await interface.dispatcher_interface.broadcast.Executor(
            target=self.depend_callable,
            event=interface.event,
            hasReferrer=True
        )

        result_is_asyncgen = [inspect.isasyncgen, isasyncgen][isinstance(result, Hashable)](result)
        result_is_generator = [inspect.isgenerator, isgenerator][isinstance(result, Hashable)](result)
        if result_is_asyncgen or (result_is_generator and not iscoroutinefunction(self.depend_callable)):
            if result_is_generator(result):
                for i in result:
                    yield Force(i)
            elif result_is_asyncgen(result):
                async for i in result:
                    yield Force(i)
        else:
            if self.cache:
                interface.local_storage[self.depend_callable] = result
            yield Force(result)
            return

class Middleware(Decorater):
    pre = True
    context_target: Any

    def __init__(self, context_target: ContextManager):
        self.context_target = context_target

    async def target(self, interface: DecoraterInterface):
        if all([
            hasattr(self.context_target, "__aenter__"),
            hasattr(self.context_target, "__aexit__")
        ]):
            async with self.context_target as mw_value:
                yield mw_value
        elif all([
            hasattr(self.context_target, "__enter__"),
            hasattr(self.context_target, "__exit__")
        ]):
            with self.context_target as mw_value:
                yield mw_value
        else:
            raise InvaildContextTarget(self.context_target, "is not vaild as a context target.")


from typing import Type
from pydantic import BaseModel
import abc
from .event import BaseEvent
from ..interfaces.dispatcher import DispatcherInterface
from .dispatcher import BaseDispatcher

class BaseRule(metaclass=abc.ABCMeta):
    target_dispatcher: BaseDispatcher

    def __init__(self, target_dispatcher: BaseDispatcher) -> None:
        self.target_dispatcher = target_dispatcher

    @abc.abstractmethod
    def check(self, event: BaseEvent, dii: DispatcherInterface) -> bool:
        pass

    class Config:
        arbitrary_types_allowed = True

class SpecialEventType(BaseRule):
    target_dispatcher: BaseDispatcher

    def __init__(self, event_type: Type[BaseEvent], target_dispatcher: BaseDispatcher, specially: bool = False) -> None:
        self.target_dispatcher = target_dispatcher
        self.event_type = event_type
        self.specially = specially

    def check(self, event: BaseEvent, dii: DispatcherInterface) -> bool:
        if self.specially:
            if type(event) is self.event_type:
                return True
        else:
            if isinstance(event, self.event_type):
                return True


from typing import Any, AsyncGenerator, Callable, Generator, List, Optional, Tuple, Type, Union, _GenericAlias

from ..entities.dispatcher import BaseDispatcher
from ..entities.event import BaseEvent
from ..entities.signatures import Force
from ..exceptions import (OutOfMaxGenerater,
                          RequirementCrashed)
from ..utilles import is_asyncgener, isgeneratorfunction, run_always_await

def get_raw_dispatcher_callable(dispatcher: Any):
  if isinstance(dispatcher, BaseDispatcher):
    return dispatcher.catch
  if callable(dispatcher):
    return dispatcher

class ContextStackItem:
  __slots__ = ("dispatchers", "event", "_index")

  dispatchers: List[BaseDispatcher]
  event: BaseEvent
  _index: int

  def __init__(self, dispatchers: List[BaseDispatcher], event: BaseEvent) -> None:
    self.dispatchers = dispatchers
    self.event = event
    self._index = 0

class ExecuteContextStackItem:
  def __init__(self, name, annotation, default, local_dispatchers, optional=False) -> None:
    self.name, self.annotation, self.default, self.local_dispatchers = \
      name, annotation, default, local_dispatchers
    self.optional = optional
    self.always_dispatchers = []
    for i in local_dispatchers:
      if isinstance(i, BaseDispatcher) and i.always:
        self.always_dispatchers.append(i)
  
  def __repr__(self) -> str:
    return "<ContextStackItem name={0} annotation={1} default={2} locald={3}"\
      .format(
        self.name, self.annotation, self.default, self.local_dispatchers)

  name: str
  annotation: Any
  default: Any
  local_dispatchers: List[Union[BaseDispatcher, Callable]]
  optional: bool

  # 即 无论如何都会被激活至少 1 次的 Dispatchers.
  always_dispatchers: List[Union[BaseDispatcher, Callable]]

  __slots__ = ("name", "annotation", "default", "local_dispatchers", "optional", "always_dispatchers")

class DispatcherInterface:
  __slots__ = ("broadcast", "execute_context_stack", "context_stack", "alive_generater_dispatcher")

  broadcast: "Broadcast"

  execute_context_stack: List[ExecuteContextStackItem]
  context_stack: List[ContextStackItem]
  alive_generater_dispatcher: List[
    List[Tuple[Union[Generator, AsyncGenerator], bool]]
    # True => async, False => sync
  ]

  @property
  def name(self):
    return self.execute_context_stack[-1].name

  @property
  def annotation(self):
    return self.execute_context_stack[-1].annotation

  @property
  def default(self):
    return self.execute_context_stack[-1].default

  @property
  def _index(self):
    return self.context_stack[-1]._index

  @_index.setter
  def _(self, new_value):
    self.context_stack[-1]._index = new_value

  @property
  def local_dispatchers(self):
    return self.execute_context_stack[-1].local_dispatchers

  @property
  def dispatchers(self):
    return self.context_stack[-1].dispatchers
  
  @property
  def event(self):
    return self.context_stack[-1].event

  # dispatcher 允许任何值, 包括 Decorater, 只不过 Decorater Dispatcher 的优先级是隐式最高, 所以基本上不可能截获到.

  def __init__(self, broadcast_instance: "Broadcast"):
    self.broadcast = broadcast_instance
    self.alive_generater_dispatcher = [[]]
    self.context_stack = []
    self.execute_context_stack =\
      [ExecuteContextStackItem(None, None, None, [])]
  
  def enter_context(self, event: BaseEvent, dispatchers: List[BaseDispatcher]):
    self.context_stack.append(ContextStackItem(dispatchers, event))
    for i in self.dispatchers[self._index:]:
      if getattr(i, "always", False):
        self.always_dispatcher.append(i)
    return self

  async def __aenter__(self) -> "DispatcherInterface":
    return self

  async def __aexit__(self, _, exc, tb):
    if len(self.alive_generater_dispatcher) != 0:
      await self.alive_dispatcher_killer()
      self.alive_generater_dispatcher.pop()
    self.context_stack.pop()

    if tb is not None:
      raise exc

  def inject_local_dispatcher(self, *dispatchers: List[Union[BaseDispatcher, Callable]]):
    self.local_dispatchers.extend(dispatchers)
    always_dispatchers = self.execute_context_stack[-1].always_dispatchers

    for i in dispatchers:
      if getattr(i, "always", False):
        always_dispatchers.append(i)

  def inject_global_dispatcher(self, *dispatchers: List[Union[BaseDispatcher, Callable]]):
    self.dispatchers.extend(dispatchers)
    always_dispatchers = self.execute_context_stack[-1].always_dispatchers

    for i in dispatchers:
      if getattr(i, "always", False):
        always_dispatchers.append(i)

  async def execute_with(self, name: str, annotation, default):
    optional = False
    if isinstance(annotation, _GenericAlias):
      if annotation.__origin__ is Union and annotation.__args__[-1] is type(None):
        # 如果是 Optional, 则它的最后一位应为 NoneType.
        optional = True
        annotation = annotation.__args__[0]
      else:
        raise TypeError("cannot parse this annotation: {0}".format(annotation))

    self.execute_context_stack.append(ExecuteContextStackItem(name, annotation, default, [], optional=optional))

    alive_dispatchers = []
    self.alive_generater_dispatcher.append(alive_dispatchers)
    
    always_dispatcher = self.execute_context_stack[-1].always_dispatchers
    
    result = None
    try:
      for self.context_stack[-1]._index, dispatcher in enumerate(
          self.dispatchers[self._index+1:],
          start=self._index+1):
        
        if getattr(dispatcher, "always", False):
          try: always_dispatcher.remove(dispatcher)
          except: pass

        if isinstance(dispatcher, type) and issubclass(dispatcher, BaseDispatcher):
          local_dispatcher = dispatcher().catch
        elif isinstance(dispatcher, BaseDispatcher):
          local_dispatcher = dispatcher.catch
        elif callable(dispatcher):
          local_dispatcher = dispatcher
        else:
          raise ValueError("invaild dispatcher: ", dispatcher)
      
        if is_asyncgener(local_dispatcher):
          now_dispatcher_generater = local_dispatcher(self).__aiter__()

          try:
            result = await now_dispatcher_generater.__anext__()
          except StopAsyncIteration:
            continue
          else:
            alive_dispatchers.append(
              (now_dispatcher_generater, True)
            )
        elif isgeneratorfunction(local_dispatcher):
          now_dispatcher_generater = local_dispatcher(self).__iter__()

          try:
            result = now_dispatcher_generater.__next__()
          except StopIteration as e:
            result = e.value
          else:
            alive_dispatchers.append(
              (now_dispatcher_generater, False)
            )
        else:
          result = await run_always_await(local_dispatcher(self))

        if result is None:
          continue
        
        if result.__class__ is Force:
          result = result.target

        self.context_stack[-1]._index = 0
        return result
      else:
        if optional:
          self.context_stack[-1]._index = 0
          return None
        raise RequirementCrashed("the dispatching requirement crashed: ", self.name, self.annotation, self.default)
    finally: # 在某种程度上, 这算 BCC 内的第二个 "无头充数" 的设计了.
      for always_dispatcher in always_dispatcher:
        if isinstance(always_dispatcher, type) and issubclass(always_dispatcher, BaseDispatcher):
          local_dispatcher = always_dispatcher().catch
        elif isinstance(always_dispatcher, BaseDispatcher):
          local_dispatcher = always_dispatcher.catch
        elif callable(always_dispatcher):
          local_dispatcher = always_dispatcher
        else:
          raise ValueError("invaild dispatcher: ", always_dispatcher)

        if is_asyncgener(local_dispatcher):
          now_dispatcher_generater = local_dispatcher(self).__aiter__()
          alive_dispatchers.append(
            (now_dispatcher_generater, True)
          )

          try:
            await now_dispatcher_generater.__anext__()
          except StopAsyncIteration:
            pass
        elif isgeneratorfunction(local_dispatcher):
          now_dispatcher_generater = local_dispatcher(self).__iter__()
          alive_dispatchers.append(
            (now_dispatcher_generater, False)
          )

          try:
            now_dispatcher_generater.__next__()
          except StopIteration as e:
            pass
        else:
          await run_always_await(local_dispatcher(self))

      self.execute_context_stack.pop()

  async def alive_dispatcher_killer(self):
    for unbound_gen, is_async_gen in self.alive_generater_dispatcher[-1]:
      index = 0
      if is_async_gen:
        async for _ in unbound_gen:
          if index == 15:
            raise OutOfMaxGenerater(
              "a dispatch as a sync generator had to stop: ", unbound_gen
            )
          index += 1
      else:
        for _ in unbound_gen:
          if index == 15:
            raise OutOfMaxGenerater(
              "a dispatch as a sync generator had to stop: ", unbound_gen
            )
          index += 1
  
  def getDispatcher(self, dispatcher_class: Type[BaseDispatcher]) -> Optional[BaseDispatcher]:
    for i in self.dispatchers:
      if type(i) is dispatcher_class:
        return i


from typing import List, NoReturn, Optional, Type, Union

from .entities.inject_rule import BaseRule
from .entities.decorater import Decorater
from .entities.dispatcher import BaseDispatcher
from .entities.event import BaseEvent
from .exceptions import InvaildEventName, RegisteredEventListener
from .entities.namespace import Namespace
from .entities.listener import Listener

class Zone:
    namespace: Namespace
    listeners: List[Listener]
    dispatcher_inject_rules: List[BaseRule]

    def __init__(self, name: str, priority: Optional[int] = None,
        injected_dispatchers: Optional[List[BaseDispatcher]] = None
    ) -> None:
        self.namespace = Namespace(
            name=name,
            priority=priority,
            injected_dispatchers=injected_dispatchers or []
        )
        self.listeners = []
        self.dispatcher_inject_rules = []
    
    @staticmethod
    def event_class_generator(target=BaseEvent):
        for i in target.__subclasses__():
          yield i
          if i.__subclasses__():
            yield from Zone.event_class_generator(i)
    
    @staticmethod
    def findEvent(name: str):
        for i in Zone.event_class_generator():
          if i.__name__ == name:
            return i
    
    def containListener(self, target):
        for i in self.listeners:
            if i.callable == target:
                return True
        return False

    def getListener(self, target):
        for i in self.listeners:
            if i.callable == target:
                return i

    def removeListener(self, target):
        self.listeners.remove(target)

    def receiver(self,
        event: Union[str, Type[BaseEvent]],
        priority: int = 16,
        dispatchers: List[Type[BaseDispatcher]] = [],
        headless_decoraters: List[Decorater] = [],
        enable_internal_access: bool = False
    ):
        if isinstance(event, str):
            _name = event
            event = self.findEvent(event)
            if not event:
                raise InvaildEventName(_name, "is not vaild!")
        priority = (type(priority) == int) and priority or int(priority) # 类型转换
        def receiver_wrapper(callable_target):
            may_listener = self.getListener(callable_target)
            if not may_listener:
                self.listeners.append(Listener(
                    callable=callable_target,
                    namespace=self.namespace,
                    inline_dispatchers=dispatchers,
                    priority=priority,
                    listening_events=[event],
                    headless_decoraters=headless_decoraters,
                    enable_internal_access=enable_internal_access
                ))
            else:
                if event not in may_listener.listening_events:
                    may_listener.listening_events.append(event)
                else:
                    raise RegisteredEventListener(event.__name__, "has been registered!")
            return callable_target
        return receiver_wrapper

    def addInjectionRule(self, rule: BaseRule) -> NoReturn:
        if rule in self.dispatcher_inject_rules:
            raise ValueError("this rule has already been added!")
        self.dispatcher_inject_rules.append(rule)
    
    def removeInjectionRule(self, rule: BaseRule) -> NoReturn:
        self.dispatcher_inject_rules.remove(rule)


class OutOfMaxGenerater(Exception):
    pass

class InvaildDispatcher(Exception):
    pass

class RequirementCrashed(Exception):
    pass

class DisabledNamespace(Exception):
    pass

class ExistedNamespace(Exception):
    pass

class UnexistedNamespace(Exception):
    pass

class RegisteredEventListener(Exception):
    pass

class InvaildEventName(Exception):
    pass

class InvaildContextTarget(Exception):
    pass

class PropagationCancelled(Exception):
    pass

class ExecutionStop(Exception):
    pass


import asyncio
import inspect
import traceback
from typing import Callable, Dict, Generator, Hashable, List, NoReturn, Optional, Type, Union
import types

from iterwrapper import IterWrapper as iw

from .builtin.event import ExceptionThrowed
from .entities.decorater import Decorater
from .entities.dispatcher import BaseDispatcher
from .entities.event import BaseEvent
from .entities.inject_rule import BaseRule
from .entities.listener import Listener
from .entities.namespace import Namespace
from .entities.signatures import Force, RemoveMe
from .exceptions import (DisabledNamespace, ExecutionStop, ExistedNamespace,
                         InvaildEventName, PropagationCancelled,
                         RegisteredEventListener, RequirementCrashed,
                         UnexistedNamespace)
from .interfaces.decorater import DecoraterInterface
from .interfaces.dispatcher import DispatcherInterface
from .utilles import (argument_signature, dispatcher_mixin_handler, group_dict, isasyncgen, isgenerator, run_always_await)
from .zone import Zone

class Broadcast:
  loop: asyncio.AbstractEventLoop

  default_namespace: Namespace
  namespaces: List[Namespace]
  listeners: List[Listener]

  # 规则驱动注入
  dispatcher_inject_rules: List[BaseRule] = []
  
  dispatcher_interface: DispatcherInterface

  debug_flag: bool

  def __init__(self, *, loop: asyncio.AbstractEventLoop = None, debug_flag: bool = False,
    inject_rules: Optional[List[BaseRule]] = None
  ):
    self.loop = loop or asyncio.get_event_loop()
    self.default_namespace = Namespace(name="default", default=True)
    self.debug_flag = debug_flag
    self.namespaces = []
    self.listeners = []
    self.dispatcher_inject_rules = inject_rules or []
    self.dispatcher_interface = DispatcherInterface(self)
  
  def default_listener_generator(self, event_class) -> Listener:
    yield from (iw(self.listeners)
      .filter(lambda x: not x.namespace.hide) # filter for hide
      .filter(lambda x: not x.namespace.disabled) # filter for disabled
      .filter(lambda x: event_class in x.listening_events)
      .collect(list) # collect to a whole list
    )

  def addInjectionRule(self, rule: BaseRule) -> NoReturn:
    if rule in self.dispatcher_inject_rules:
      raise ValueError("this rule has already been added!")
    self.dispatcher_inject_rules.append(rule)
  
  def removeInjectionRule(self, rule: BaseRule) -> NoReturn:
    self.dispatcher_inject_rules.remove(rule)

  async def layered_scheduler(self,
    listener_generator: Generator[Listener, None, None],
    event: BaseEvent
  ):
    grouped: Dict[int, List[Listener]] = group_dict(listener_generator, lambda x: x.priority)
    break_flag = False
    for current_priority in sorted(grouped.keys()):
      if break_flag:
        break
      current_group = grouped[current_priority]
      try:
        await asyncio.gather(*[
          self.Executor(
            target=i,
            event=event
          ) for i in current_group
        ])
      except Exception as e:
        if isinstance(e, PropagationCancelled):
          break

  async def Executor(self,
    target: Union[Callable, Listener],
    event: BaseEvent,
    dispatchers: List[Union[
      Type[BaseDispatcher],
      Callable,
      BaseDispatcher,
    ]] = None,
    hasReferrer: bool = False,
    enableInternalAccess: bool = False
  ):
    is_listener = isinstance(target, Listener)

    if is_listener:
      if target.namespace.disabled:
        raise DisabledNamespace("catched a disabled namespace: {0}".format(target.namespace.name))

    # 先生成 dispatchers
    _dispatchers = dispatcher_mixin_handler(event.Dispatcher)
    # 开始暴力注入
    if is_listener:
      if target.inline_dispatchers:
        _dispatchers = target.inline_dispatchers + _dispatchers
      if target.namespace.injected_dispatchers:
        _dispatchers = target.namespace.injected_dispatchers + _dispatchers
    if dispatchers:
      _dispatchers = dispatchers + _dispatchers

    target_callable = target.callable if is_listener else target
    parameter_compile_result = {}

    async with self.dispatcher_interface.enter_context(event, _dispatchers) as dii:
      if (not dii.dispatchers) or type(dii.dispatchers[0]) is not DecoraterInterface:
        DecoraterInterface(dii)  # pylint: disable=unused-variable
      # Decorater 的 Dispatcher 已经注入, 没他事了

      if enableInternalAccess or \
        (is_listener and \
          target.enable_internal_access):
        internal_access_mapping = {
          Broadcast: self,
          DispatcherInterface: dii,
          **({ # 当 protocol.target 为 Listener 时的特有解析
            Listener: target,
            Namespace: target.namespace
          } if is_listener else {})
        }
        @dii.inject_global_dispatcher
        def _(interface: DispatcherInterface):
          return internal_access_mapping.get(interface.annotation)

      for injection_rule in self.dispatcher_inject_rules:
        if injection_rule.check(event, dii):
          dii.dispatchers.insert(1, injection_rule.target_dispatcher)

      @dii.inject_global_dispatcher
      def _(interface: DispatcherInterface):
        if interface.annotation is event.__class__:
          return event

      try:
        for name, annotation, default in argument_signature(target_callable):
          parameter_compile_result[name] =\
            await dii.execute_with(name, annotation, default)
        if is_listener:
          if target.headless_decoraters: # 无头装饰器
            for hl_d in target.headless_decoraters:
              await dii.execute_with(None, None, hl_d)
      except ExecutionStop:
        raise
      except RequirementCrashed:
        traceback.print_exc()
        raise
      except Exception as e:
        traceback.print_exc()
        if not hasReferrer: # 如果没有referrer, 会广播事件, 如果有则向上抛出
          self.postEvent(ExceptionThrowed(
            exception=e,
            event=event
          ))
        raise

      try:
        result = await run_always_await(
          target_callable(**parameter_compile_result)
        )
      except ExecutionStop:
        raise # 直接抛出.
      except PropagationCancelled:
        raise # 防止事件广播
      except Exception as e:
        traceback.print_exc()
        if not hasReferrer: # 如果没有referrer, 则广播事件, 如果有则向上抛出(防止重复抛出事件)
          self.postEvent(ExceptionThrowed(
            exception=e,
            event=event
          ))
        raise
      
      if [inspect.isgenerator, isgenerator][isinstance(result, Hashable)](result):
        try: result = next(result)
        except StopIteration as e: result = e.value
        else: dii.alive_generater_dispatcher[-1].append(
          (result, False)
        )
      elif [inspect.isasyncgen, isasyncgen][isinstance(result, Hashable)](result):
        try: result = await result.__anext__()
        except StopAsyncIteration: return # value = None
        else: dii.alive_generater_dispatcher[-1].append(
          (result, True)
        )

      if result.__class__ is Force:
        return result.content

      if result.__class__ is RemoveMe:
        if is_listener:
          if target in self.listeners:
            self.listeners.pop(self.listeners.index(target))
      return result

  def postEvent(self, event: BaseEvent):
    self.loop.create_task(self.layered_scheduler(
      listener_generator=self.default_listener_generator(event.__class__),
      event=event
    ))

  @staticmethod
  def event_class_generator(target=BaseEvent):
    for i in target.__subclasses__():
      yield i
      if i.__subclasses__():
        yield from Broadcast.event_class_generator(i)

  @staticmethod
  def findEvent(name: str):
    for i in Broadcast.event_class_generator():
      if i.__name__ == name:
        return i

  def getDefaultNamespace(self):
    return self.default_namespace
  
  def createNamespace(self, name, *, priority: int = 0, hide: bool = False, disabled: bool = False):
    if self.containNamespace(name):
      raise ExistedNamespace(name, "has been created!")
    self.namespaces.append(Namespace(name=name, priority=priority, hide=hide, disabled=disabled))
    return self.namespaces[-1]
  
  def removeNamespace(self, name):
    if self.containNamespace(name):
      for index, i in enumerate(self.namespaces):
        if i.name == name:
          self.namespaces.pop(index)
          return
    else:
      raise UnexistedNamespace(name)

  def containNamespace(self, name):
    for i in self.namespaces:
      if i.name == name:
        return True
    return False
  
  def getNamespace(self, name):
    if self.containNamespace(name):
      for i in self.namespaces:
        if i.name == name:
          return i
    else:
      raise UnexistedNamespace(name)

  def hideNamespace(self, name):
    ns = self.getNamespace(name)
    ns.hide = True

  def unhideNamespace(self, name):
    ns = self.getNamespace(name)
    ns.hide = False

  def disableNamespace(self, name):
    ns = self.getNamespace(name)
    ns.disabled = True

  def enableNamespace(self, name):
    ns = self.getNamespace(name)
    ns.disabled = False

  def containListener(self, target):
    for i in self.listeners:
      if i.callable == target:
        return True
    return False

  def getListener(self, target):
    for i in self.listeners:
      if i.callable == target:
        return i

  def removeListener(self, target):
    self.listeners.remove(target)

  def includeZone(self, target_zone: Zone):
    if self.containNamespace(Zone.namespace.name):
      raise ValueError("the zone cannot import because its name has been used.")
    self.namespaces.append(target_zone.namespace)
    self.listeners.extend(target_zone.listeners)

  def receiver(self,
      event: Union[str, Type[BaseEvent]],
      priority: int = 16,
      dispatchers: List[Type[BaseDispatcher]] = [],
      namespace: Namespace = None,
      headless_decoraters: List[Decorater] = [],
      enable_internal_access: bool = False
  ):
    if isinstance(event, str):
      _name = event
      event = self.findEvent(event)
      if not event:
        raise InvaildEventName(_name, "is not vaild!")
    priority = (type(priority) == int) and priority or int(priority) # 类型转换
    def receiver_wrapper(callable_target):
      may_listener = self.getListener(callable_target)
      if not may_listener:
        self.listeners.append(Listener(
          callable=callable_target,
          namespace=namespace or self.getDefaultNamespace(),
          inline_dispatchers=dispatchers,
          priority=priority,
          listening_events=[event],
          headless_decoraters=headless_decoraters,
          enable_internal_access=enable_internal_access
        ))
      else:
        if event not in may_listener.listening_events:
          may_listener.listening_events.append(event)
        else:
          raise RegisteredEventListener(event.__name__, "has been registered!")
      return callable_target
    return receiver_wrapper



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['graia',
 'graia.broadcast',
 'graia.broadcast.builtin',
 'graia.broadcast.entities',
 'graia.broadcast.entities.signatures',
 'graia.broadcast.interfaces',
 'graia.broadcast.protocols']

package_data = \
{'': ['*']}

install_requires = \
['iterwrapper>=0.1.2,<0.2.0', 'pydantic>=1.5.1,<2.0.0']

setup_kwargs = {
    'name': 'graia-broadcast',
    'version': '0.2.2',
    'description': 'a highly customizable, elegantly designed event system based on asyncio',
    'long_description': None,
    'author': 'Chenwe_i_lin',
    'author_email': 'Chenwe_i_lin@outlook.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from typing import Callable, List, Type

from .dispatcher import BaseDispatcher
from .event import BaseEvent
from .namespace import Namespace
from .decorater import Decorater

class Listener:
    def __init__(self,
        callable: Callable,
        namespace: Namespace,
        listening_events: List[Type[BaseEvent]],
        inline_dispatchers: List[BaseDispatcher] = None,
        headless_decoraters: List[Decorater] = None,
        priority: int = 16,
        enable_internal_access: bool = False,
    ) -> None:
        self.callable = callable
        self.namespace = namespace
        self.listening_events = listening_events
        self.inline_dispatchers = inline_dispatchers or []
        self.headless_decoraters = headless_decoraters or []
        self.priority = priority
        self.enable_internal_access = enable_internal_access

    callable: Callable
    namespace: Namespace
    listening_events: List[Type[BaseEvent]]
    inline_dispatchers: List[BaseDispatcher] = []
    headless_decoraters: List[Decorater] = []
    priority: int = 16
    enable_internal_access: bool = False


import inspect
from typing import Any, Callable, List

from iterwrapper import IterWrapper
from functools import lru_cache
from .entities.dispatcher import BaseDispatcher

async def run_always_await(any_callable):
    if inspect.iscoroutine(any_callable):
        return await any_callable
    else:
        return any_callable

def printer(value):
    print(value)
    return value

def group_dict(iw: IterWrapper, key: Callable[[Any], Any]):
    temp = {}
    for i in iw:
        k = key(i)
        temp.setdefault(k, [])
        temp[k].append(i)
    return temp

@lru_cache(None)
def argument_signature(callable_target):
    return [
        (name,
        param.annotation if param.annotation != inspect._empty else None,
        param.default if param.default != inspect._empty else None)
        for name, param in dict(inspect.signature(callable_target).parameters).items()
    ]

cache_size = None

@lru_cache(cache_size)
def is_asyncgener(o):
    return inspect.isasyncgenfunction(o)

@lru_cache(cache_size)
def isgeneratorfunction(o):
    return inspect.isgeneratorfunction(o)

@lru_cache(cache_size)
def iscoroutinefunction(o):
    return inspect.iscoroutinefunction(o)

@lru_cache(cache_size)
def isasyncgen(o):
    return inspect.isasyncgen(o)

@lru_cache(cache_size)
def isgenerator(o):
    return inspect.isgenerator(o)

async def whatever_gen_once(any_gen, *args, **kwargs):
    if inspect.isasyncgenfunction(any_gen):
        # 如果是异步生成器函数
        async for i in any_gen(*args, **kwargs):
            return i
    elif (inspect.isgeneratorfunction(any_gen) and \
        not inspect.iscoroutinefunction(any_gen)):
        # 同步生成器
        for i in any_gen(*args, **kwargs):
            return i
    elif inspect.isgenerator(any_gen):
        for i in any_gen:
            return i
    elif inspect.isasyncgen(any_gen):
        async for i in any_gen:
            return i

def flat_yield_from(l):
    for i in l:
        if type(i) == list:
            yield from i

def dispatcher_mixin_handler(dispatcher: BaseDispatcher) -> List[BaseDispatcher]:
    unbound_mixin = getattr(dispatcher, "mixin", [])
    result = [dispatcher]

    BaseDispatcher.__bases__
    for i in unbound_mixin:
        if issubclass(i, BaseDispatcher):
            result.extend(dispatcher_mixin_handler(i))
        else:
            result.append(i)
    return result


from typing import List

from pydantic import BaseModel  # pylint: disable=no-name-in-module

from .dispatcher import BaseDispatcher


class Namespace(BaseModel):
    name: str
    injected_dispatchers: List[BaseDispatcher] = []

    priority: int = 0
    default: bool = False
    
    hide: bool = False
    disabled: bool = False

    class Config:
        arbitrary_types_allowed = True
